/*jslint node:true*/

(function () {
    'use strict';

    var async = require('async'),
        getEnv = require('./commons').getEnv,
        _ = require('underscore'),
        path = require('path'),
        grunt = require('grunt'),
        minifyBatch = require('./minify-batch');

    module.exports = {
        'pack-lib': {
            description: 'Create Packages for JavaScript and CSS files from the specified profile.',
            args: [{
                alias: ['-s', '--store-type'],
                nargs: 1,
                required: true,
                help: 'Store type e.g. Store'
            }, {
                alias: ['-v', '--version'],
                nargs: 1,
                required: false,
                help: 'Store version e.g. 6.17.0'
            }, {
                alias: ['-w', '--webroot'],
                nargs: 1,
                required: false,
                help: 'Epages webroot defaults to $EPAGES_WEBROOT'
            }, {
                alias: ['-t', '--package-type'],
                nargs: 1,
                required: false,
                choices: ['less', 'js'],
                help: 'Type of package that should be built'
            },
            {
                alias: ['-p', '--profile'],
                action: 'append',
                nargs: 1,
                required: false,
                choices: ['bo', 'sf', 'mobilesf', 'sf_appointment', 'pageeditor', 'editor', 'testsuite'],
                help: 'Profile name of resulting package'
            }],
            example: 'Example: epnode pack --js --css -p sf -p mobilesf -p editor -p testsuite -t Store -v 6.17.0',
            exec: function (args, callback) {
                var pack = require('./pack'),
                    vars = ['webroot', 'log'];

                // Try to get version if non was specified.
                if (!args.version) { vars.push('version'); }

                getEnv(vars, function (err, data) {
                    if (err) { callback(err); }

                    // mixin environment data
                    _.extend(data, args);
                    switch (data.package_type) {
                    case 'js':
                        pack.js(data, callback);
                        break;
                    case 'less':
                        pack.less(data, callback);
                        break;
                    //pack less and js for all profiles
                    default:
                        async.parallel([
                            function (callback) {
                                pack.js(data, callback);
                            },
                            function (callback) {
                                pack.less(data, callback);
                            }
                        ], callback);
                    }
                });
            }
        },
        'minify-lib': {
            description: 'Minify all JavaScript and CSS files within the lib folder.',
            args: [{

                alias: ['-s', '--store-type'],
                nargs: 1,
                required: true,
                help: 'Store type e.g. Store'
            }, {

                alias: ['-v', '--version'],
                nargs: 1,
                required: false,
                help: 'Store version e.g. 6.17.0'
            }, {

                alias: ['-w', '--webroot'],
                nargs: 1,
                required: false,
                help: 'Epages webroot defaults to $EPAGES_WEBROOT'
            }, {

                alias: ['-f', '--filter'],
                nargs: 1,
                required: false,
                help: 'Filter files to be minified e.g. /**/*.js'
            }],
            example: 'Example: epnode minify -t Store -v 6.17.0 -f /**/*.ca.json',
            exec: function (args, callback) {
                var vars = ['webroot', 'log'];

                // Try to get version if non was specified.
                if (!args.version) { vars.push('version'); }

                // mixin environment data
                getEnv(vars, function (err, data) {
                    if (err) { callback(err); }

                    _.extend(data, args);
                    var dir = path.join(data.webroot, 'StoreTypes', data.version, data.store_type),
                        source = path.join(dir, 'lib'),
                        target = path.join(dir, 'lib-min');

                    if (grunt.file.isDir(source)) {
                        minifyBatch(source, target, {'uid': data.uid, 'gid': data.gid, 'filter': data.filter}, callback);
                    } else {
                        callback('Cannot find files to minify at ' + dir);
                    }
                });

            }
        },
        'minify': {
            description: 'Minify the specified source file or folder to the specified target.',
            args: [{
                alias: ['-s', '--source'],
                nargs: 1,
                required: true,
                help: 'Source that should be minified. Can either be a file or a folder.'
            }, {
                alias: ['-t', '--target'],
                nargs: 1,
                required: true,
                help: 'The target for the minfied file or folder. If <source> is a folder, target must be a folder too.'
            }, {
                alias: ['-f', '--filter'],
                nargs: 1,
                required: false,
                help: 'Filter files to be minified e.g. /**/*.js'
            }],
            example: 'Example: epnode minify -s lib -t lib-min',
            exec: function (args, callback) {

                // mixin environment data
                getEnv(['webroot', 'log'], function (err, data) {
                    if (err) { callback(err); }

                    _.extend(data, args);
                    // sFile => tFile (tFile is created as a minified version of sFile)
                    // sDir => tDir (All files in sDir will be minified into tDir)
                    var is_fileToFile = (grunt.file.isFile(data.source) && !grunt.file.isDir(data.target)),
                        is_dirToDir = (grunt.file.isDir(data.source) && !grunt.file.isFile(data.target));

                    if (data.source === data.target) {
                        callback('Cannot replace file with minified version of itself (yet). SOURCE is identical to TARGET');
                    } else if (is_fileToFile || is_dirToDir) {
                        minifyBatch(data.source, data.target, {'uid': data.uid, 'gid': data.gid, 'filter': data.filter}, callback);
                    } else {
                        callback('Source not found or mixed files and directories. Use: sFile => tFile || sDir => tDir');
                    }
                });
            }
        },
        'build-ui': {
            description: 'Create Packages for JavaScript and CSS files for the specified store type.',
            args: [{
                alias: ['-s', '--store-type'],
                nargs: 1,
                required: true,
                help: 'Store type e.g. Store'
            }, {
                alias: ['-v', '--version'],
                nargs: 1,
                required: false,
                help: 'Store version e.g. 6.17.0'
            }, {
                alias: ['-w', '--webroot'],
                nargs: 1,
                required: false,
                help: 'Epages webroot defaults to $EPAGES_WEBROOT'
            }, {
                alias: ['-f', '--filter'],
                nargs: 1,
                required: false,
                help: 'Filter files to be minified e.g. /**/*.js'
            }],
            example: 'Example: epnode build-ui -s Store',
            exec: function (args, callback) {
                var pack = require('./pack'),
                    vars = ['webroot', 'log'];

                // Try to get version if non was specified.
                if (!args.version) { vars.push('version'); }

                getEnv(vars, function (err, data) {
                    if (err) { callback(err); }

                    _.extend(data, args);

                    async.auto({
                        packJs: function (callback) {
                            pack.js(data, callback);
                        },

                        packLess: function (callback) {
                            pack.less(data, callback);
                        },

                        minify: ['packJs', 'packLess', function (callback) {
                            var dir = path.join(data.webroot, 'StoreTypes', data.version, data.store_type),
                                source = path.join(dir, 'lib'),
                                target = path.join(dir, 'lib-min');

                            minifyBatch(source, target, {'uid': data.uid, 'gid': data.gid, 'filter': data.minify_filter}, callback);
                        }]
                    }, callback);
                });
            }
        },
        'testcases': {
            description: 'Check for the presence of testcases.',
            args: [{
                alias: ['-c', '--cartridge'],
                nargs: 1,
                required: false,
                action: 'append',
                help: 'Check if the cartridge has testcases [multi]',
                example: 'Example: epnode testcases -c jquery -c presentation'

            }],
            example: 'Example: epnode testcases -c jquery',
            exec: function (args, callback) {
                var testcases = require('./testcases');

                testcases(args.cartridge, function (error) {
                    if (!error) {
                        console.log('Seems like every module is covered by at leat one testcase!');
                        callback();
                    } else {
                        // write buffer for erro if necessary
                        callback(error);

                    }
                });
            }
        },
        'builddoc': {
            description: 'Create the jsduck documentation.',
            args: [{
                alias: ['-i', '--include'],
                nargs: 1,
                required: true,
                help: 'A JSON file which enumerates the files to include in the documentation.'
            }, {
                alias: ['-o', '--output'],
                nargs: 1,
                required: false,
                help: 'The output folder for the documentation'
            }, {
                alias: ['-c', '--cartridges'],
                nargs: 1,
                required: false,
                help: 'The folder containing the cartridges, typically /srv/epages/eproot/Cartridges'
            }],
            example: 'Example: epnode builddoc -i sources.json -o ./',
            exec: function (args, callback) {
                var builddoc = require('./builddoc');

                getEnv(['log', 'cartridges', 'webroot'], function (data) {
                    _.extend(data, args);

                    builddoc({
                        'input': data.include,
                        'output': data.output || path.join(data.webroot, 'Docs', 'JavaScript'),
                        'cartridges': data.cartridges,
                        'uid': data.uid,
                        'gid': data.gid
                    }, function (error) {
                        if (error) {
                            console.log('[ERROR] Build documentation');
                        }

                        callback(error);
                    });
                });
            }
        }
    };
}());