/*jslint indent:4,node:true,stupid:true*/

(function () {
    'use strict';

    var _ = require('underscore'),
        path = require('path'),
        grunt = require('grunt'),
        fs = require('fs'),
        perl = require('./perl'),
        os = require('os'),
        async = require('async'),
        getEnv = function (vars, callback) {
            var env = {};

            async.parallel([
                function (callback) {
                    perl.get('env', vars, function (err, data) {
                        _.each(data, function (val, ix) { env[ix] = val; });

                        callback(err);
                    });
                },

                function (callback) {
                    if (os.platform() !== 'win32') {
                        perl.get('uid', 'appuser', function (appuser_id) {
                            env.uid = appuser_id;
                            callback(null);
                        });
                    } else {
                        callback(null);
                    }
                },

                function (callback) {
                    if (os.platform() !== 'win32') {
                        perl.get('gid', 'webgroup', function (webgroup_id) {
                            env.gid = webgroup_id;
                            callback(null);
                        });
                    } else {
                        callback(null);
                    }
                }
            ], function (error) {
                callback(error, env);
            });
        },

        buildTemplateModule = function (module, source) {
            var html,
                wrapper;

            html = fs.readFileSync(source).toString();

            // This is the JS code used to wrap a template. `##MODULE_NAME##` is replaced by the
            // template name and `##HTML##` is replaced by the actual content of the template.
            wrapper = 'define("$tmpl!##MODULE_NAME##", ["jquery/tmpl"], function ($) {' + '    $.template("##TEMPLATE_NAME_ONE##", "##HTML##");' + '    var tmpl = $.template("##TEMPLATE_NAME_TWO##", "##HTML##");' + '    return function (data, options) {' + '        return $.tmpl(tmpl, data, options);' + '    };' + '});';

            // The double quotes and line breaks must be escaped, everything else can be left
            // unchanged.
            html = html.replace(/"/g, '\\"')
                .replace(/[\n\r]\s*/g, '');

            return wrapper
                .replace('##MODULE_NAME##', module)
                .replace('##TEMPLATE_NAME_ONE##', module.replace(/\//g, '.'))
                .replace('##TEMPLATE_NAME_TWO##', module)
                .replace(/##HTML##/g, html);
        },

        fixPath = function (p) {
            return p.replace(new RegExp(path.sep === '\\' ? '\\\\' : path.sep, 'g'), '/');
        },

        compileTemplateFiles = function (base, source, writeToDisk) {
            var dir = path.dirname(source),
                file = path.basename(source, '.js'),
                concatenatedContent = '',
                includedFiles = [];

            base = base.replace(/\\/g, '/');
            if (base.charAt(base.length - 1) !== '/') {
                base += '/';
            }

            grunt.file.expand(path.join(dir, file + '*.tmpl.html')).forEach(function (templateFile) {
                var module = templateFile.replace(base, '').replace('.tmpl.html', ''),
                    templateContent,
                    compiledTemplateFile;

                // Fix paths on Windows.
                module = fixPath(path.relative(base, templateFile).replace(/\.tmpl\.html$/, ''));

                // The actual content of the template file is decorated with some
                // boilerplate code by this function.
                templateContent = buildTemplateModule(module, templateFile);

                compiledTemplateFile = path.join(dir, path.basename(templateFile, '.html') + '.js');
                concatenatedContent += templateContent;

                if (writeToDisk) {
                    includedFiles.push(compiledTemplateFile);
                    fs.writeFileSync(compiledTemplateFile, templateContent);
                }
            });

            includedFiles.push(source);
            concatenatedContent += fs.readFileSync(source);

            return {
                includedFiles: includedFiles,
                concatenatedContent: concatenatedContent
            };
        },

        cleanUp = function (dir) {
            // remove compiled templates
            grunt.file.expand(path.join(dir, '**/*.tmpl.js')).forEach(function (compiledTemplate) {
                grunt.file.delete(compiledTemplate, {force: true});
            });
        };

    module.exports = {
        compileTemplateFiles: compileTemplateFiles,
        fixPath: fixPath,
        cleanUp: cleanUp,
        getEnv: getEnv,
        error: {
            '-1': 'unknown error',
            '0': 'success',
            '1': 'end of file',
            '2': 'getaddrinfo error',
            '3': 'permission denied',
            '4': 'resource temporarily unavailable',
            '5': 'address already in use',
            '6': 'address not available',
            '7': 'address family not supported',
            '8': 'connection already in progress',
            '9': 'bad file descriptor',
            '10': 'resource busy or locked',
            '11': 'software caused connection abort',
            '12': 'connection refused',
            '13': 'connection reset by peer',
            '14': 'destination address required',
            '15': 'bad address in system call argument',
            '16': 'host is unreachable',
            '17': 'interrupted system call',
            '18': 'invalid argument',
            '19': 'socket is already connected',
            '20': 'too many open files',
            '21': 'message too long',
            '22': 'network is down',
            '23': 'network is unreachable',
            '24': 'file table overflow',
            '25': 'no buffer space available',
            '26': 'not enough memory',
            '27': 'not a directory',
            '28': 'illegal operation on a directory',
            '29': 'machine is not on the network',
            '31': 'socket is not connected',
            '32': 'socket operation on non-socket',
            '33': 'operation not supported on socket',
            '34': 'no such file or directory',
            '35': 'function not implemented',
            '36': 'broken pipe',
            '37': 'protocol error',
            '38': 'protocol not supported',
            '39': 'protocol wrong type for socket',
            '40': 'connection timed out',
            '41': 'invalid Unicode character',
            '42': 'address family for hostname not supported',
            '44': 'servname not supported for ai_socktype',
            '45': 'ai_socktype not supported',
            '46': 'cannot send after transport endpoint shutdown',
            '47': 'file already exists',
            '48': 'no such process',
            '49': 'name too long',
            '50': 'operation not permitted',
            '51': 'too many symbolic links encountered',
            '52': 'cross-device link not permitted',
            '53': 'directory not empty',
            '54': 'no space left on device',
            '55': 'i/o error',
            '56': 'read-only file system',
            '57': 'no such device',
            '58': 'invalid seek',
            '59': 'operation canceled'
        }
    };
}());