/*jslint indent:4,node:true,nomen:true,stupid:true,regexp:true*/
(function () {
    'use strict';

    var _ = require('underscore'),
        fs = require('fs.extra'),
        os = require('os'),
        grunt = require('grunt'),
        path = require('path'),
        async = require('async'),
        chownr = require('chownr'),
        cp = require('child_process'),
        commons = require('./commons'),
        log = require('./log'),
        logger;

    log.add('File', {
        level: log.levels[log.getLogLevel(process.argv)],
        timestamp: function () {
            return new Date().getTime();
        },
        colorize: true,
        filename: path.join(process.env.EPAGES_LOG || "", 'epnode6-minify.log')
    });
    logger = log.getLogger({
        module: 'lib/minify-batch.js'
    });

    // Exports minify method
    module.exports = function (source, target, options, callback) {
        var cpuCount = os.cpus().length,
            workerLimit = cpuCount <= 3 ? 2 : Math.ceil(cpuCount / 2),
            workerMemLimit = os.freemem() / 1024 / 1024 / 20,
            workers = [],
            i,
            q,
            preFilter;

        // Since each forked node process consumes at least 10-20MB we limit the number of workers
        workerLimit = workerMemLimit < workerLimit ? workerMemLimit : workerLimit;

        // We minify directly if there is only one file. No need to call for a worker
        if (grunt.file.isFile(source)) {
            require('./minify')(path.dirname(source), source, target, {
                'id': -1
            }, function (e) {
                if (_.isNumber(options.uid) && _.isNumber(options.gid)) {
                    fs.chown(target, options.uid, options.gid, function (err) {
                        commons.cleanUp(source);
                        callback(e || err);
                    });
                } else {
                    callback(e);
                }
            });
        } else {
            options = _.extend({
                ignore: /(\.tmpl\.html|\.tmpl\.js|-test|-docu|CVS)/
            }, options);

            // Let's get some workers
            for (i = 0; i < workerLimit; i += 1) {
                workers.push({
                    id: i,
                    process: cp.fork(__dirname + '/minify-worker', [].concat(process.argv)),
                    free: true
                });
            }

            logger.info(workers.length + ' workers created (workerLimit: ' + (cpuCount <= 3 ? 2 : Math.ceil(cpuCount / 2)) + ', workerMemLimit:' + workerMemLimit + ')');

            // Setup a asynchronous queue
            // TODO: may be we can implement it without worker - as command line call (concurrent in grunt)
            q = async.queue(function (file, callback) {
                async.select(workers, function (worker, cb) {
                    cb(worker.free);
                }, function (freeWorker) {
                    // take first worker that is free
                    var worker = freeWorker[0],
                        targetFilePath = path.join(target, path.relative(source, file));

                    worker.free = false;

                    logger.debug('worker busy (id: ' + worker.id + '), file:' + file);

                    worker.process.send({
                        id: worker.id,
                        base: source,
                        source: file,
                        target: targetFilePath
                    });

                    worker.process.once('message', function (e) {
                        worker.free = true;

                        logger.debug('worker free (id: ' + worker.id + '), (task done), error: ' + e);

                        callback(e);
                    });
                });
            }, workerLimit);

            // Clean up source files created during minify process and set ownership when queue is depleted
            q.drain = function (e) {
                commons.cleanUp(source);

                if (_.isNumber(options.uid) && _.isNumber(options.gid)) {
                    chownr(target, options.uid, options.gid, callback);
                } else {
                    callback(e);
                }
            };

            // Apply pre filter for and directories and ignored files
            preFilter = {
                filter: function (path) {
                    return !options.ignore.test(path) && grunt.file.isFile(path);
                }
            };

            // Strip ignored files, apply filter and push result to the queue
            q.push(grunt.file.expand(preFilter, source + '/' + (options.filter || '**/*')));
        }
    };
}());
