/*jslint nomen:true,regexp:true,node:true,unparam:true*/
(function () {

    'use strict';

    var commons = require('./commons'),
        uglify = require('uglify-js'),
        grunt = require('grunt'),
        path = require('path'),
        html = require('html-minifier'),
        less = require('less'),
        async = require('async'),
        fs = require('fs'),
        log = require('./log'),
        logger = log.getLogger({
            module: 'lib/minify.js'
        }),
        r_xr_____ = 416,
        /*r-xr-----*/
        /**
         * Resolve all imports of the file [source] and concatenate the imported files. The imports
         * are resolved recursively, which means that the resulting content must not contain any
         * relative imports.
         *
         * Absolute imports are kept (http://example/style.css and /style.css).
         *
         * @param  {String} source The absolute path to the source file.
         * @return {String}        The content of the source file, containing the resolved imports.
         */
        resolveImports = function (source, base) {
            var content = grunt.file.read(source),
                importRegExp = /@import\s+url\s*\(\s*(["'])?(.*?)\1?\s*\);/g,
                excludeUrlRegExp = /^(data:|[a-z]+:\/\/|\/)/;

            content = content.replace(/(url\s*\(\s*(["']?))(.*?)(\2\s*\))/g, function (fullMatch, beforeUrl, c, url, afterUrl) {

                if (excludeUrlRegExp.test(url)) {
                    // we don't want to modify absolute url, hence we replace the match with itself
                    return fullMatch;
                }

                // make the url relative to the given base directory
                return beforeUrl + commons.fixPath(path.join(path.relative(base, path.dirname(source)), url)) + afterUrl;
            });

            // Search for further imports within the content of the imported file
            (content.match(importRegExp) || []).forEach(function (importUrl) {
                var importSource = importUrl.replace(importRegExp, '$2');

                if (!excludeUrlRegExp.test(importSource)) {
                    // `importSource` is relative to `base`, because we changed it in the previous
                    // step together with all other urls.
                    importSource = path.join(base, importSource);
                    content = content.replace(importUrl, resolveImports(importSource, base));
                }
            });

            return content;
        };

    // Copy / minify files by type
    module.exports = function (base, source, target, options, callback) {
        var ext = path.extname(source),
            sourceDirName = path.dirname(source),
            targetDirName = path.dirname(target),
            targetFileName = path.basename(target, ext),
            mapFileTarget = targetFileName + ext + '.map',
            content,
            parser,
            minified,
            includedFiles,
            fnCallback = callback,
            workerId = options.id || undefined;

        callback = function (err) {
            logger.debug('finished minifying (id: ' + workerId + '), (' + ext + '): ' + target);

            fnCallback(err);
        };

        base += (base.slice(-1) !== '/') ? '/' : '';

        logger.debug('starting minifying (id: ' + workerId + '): ' + targetFileName);

        switch (ext) {
        case '.js':
            // contains js and compiled templates
            includedFiles = commons.compileTemplateFiles(base, source, true).includedFiles;

            // Minify Javascript
            minified = uglify.minify(includedFiles, {
                'outSourceMap': targetFileName + ext
            });

            async.parallel([
                // Write the minified source.
                function (callback) {
                    minified.code += '\n//@ sourceMappingURL=' + mapFileTarget;
                    fs.writeFile(target, minified.code, {
                        mode: r_xr_____
                    }, callback);
                },

                // Modify and write source map.
                function (callback) {
                    minified.map = JSON.parse(minified.map);

                    minified.map.sources.forEach(function (element, ix) {
                        minified.map.sources[ix] = commons.fixPath(path.relative(targetDirName, path.join(sourceDirName, path.basename(element))));
                    });

                    fs.writeFile(path.join(targetDirName, mapFileTarget), JSON.stringify(minified.map), {
                        mode: r_xr_____
                    }, callback);
                }
            ], callback);
            break;
        case '.css':
        case '.less':
            // There is no debug level 1 for less/css files, so we're just minifying the actual packages
            if (/^package-[a-z]+/.test(path.basename(source))) {
                content = grunt.file.read(source);

                parser = new(less.Parser)({
                    'paths': [path.dirname(source)],
                    'filename': source
                });

                content = resolveImports(source, base);

                parser.parse(content, function (error, tree) {
                    if (error) {
                        throw error;
                    }

                    // Ensure package's file extension to be .css (since we processed .css AND .less files)
                    target = path.join(path.dirname(target), path.basename(target)).replace(/\.(less|css)$/, '.css');

                    fs.writeFile(target, tree.toCSS({
                        compress: true
                    }), {
                        mode: r_xr_____
                    }, callback);
                });
            } else {
                callback();
            }
            break;
        case '.html':
            fs.writeFile(target, grunt.file.read(source).replace(/[\n\r]\s*/g, ''), {
                mode: r_xr_____
            }, callback);
            break;
        case '.content':
            // replace escaped quotes by encoded html entities because gadgets seem to have a problem with these
            fs.writeFile(target, html.minify(grunt.file.read(source), {
                removeComments: true,
                collapseWhitespace: true,
                useShortDoctype: true
            }).replace(/\\"/g, '&quot;'), {
                mode: r_xr_____
            }, callback);
            break;
        default: // just copy
            grunt.file.copy(source, target);
            fs.chmod(target, r_xr_____, callback);
        }
    };
}());
