/*jslint node: true*/

(function () {
    'use strict';

    var _ = require('underscore'),
        async = require('async'),
        commons = require('./commons'),
        grunt = require('grunt'),
        fs = require('fs.extra'),
        path = require('path'),
        perl = require('./perl'),
        chownr = require('chownr'),
        targetPath,
        r_xr_____ = 416, /*r-xr-----*/

        // get an array of all cartridges for the given store type
        getCartridgeData = function (storeType, callback) {
            async.series({
                path: function (callback) {
                    perl.get('env', ['cartridges'], callback);
                },
                cartridges: function (callback) {
                    perl.get('cartridges', storeType, callback);
                }
            }, function (e, result) {
                callback(e, result.path.cartridges, result.cartridges);
            });
        },

        // set the target path where our packages will be stored,
        // e.g.: C:\epages\Shared\WebRoot\StoreTypes\6.17.3\Store\lib
        setTargetPath = function (webroot, version, type) {
            if (!targetPath) {
                targetPath = path.join(webroot, 'StoreTypes', version, type, 'lib');
            }
        };

    // create packages for javascript files
    module.exports.js = function (config, callback) {
        var scriptPackages,
            concatPackages = {},

            prepare = function (callback) {
                setTargetPath(config.webroot, config.version, config.store_type);

                // use all available stylePackages if no certain profile(s) is/are given
                scriptPackages = config.profile ? config.profile.split(' ') : ['bo', 'sf', 'mobilesf', 'sf_appointment', 'testsuite', 'editor'];

                // Set cartridge path and pass all cartridges to get_script_packages via waterfall callback
                getCartridgeData(config.store_type, callback);
            },

            get_script_packages = function (cartridgePath, cartridges, callback) {
                var t_packages = [];

                scriptPackages.forEach(function (profile) {
                    // define single methods for getting the profiles for each profile type
                    t_packages.push(function (callback) {
                        concatPackages[profile] = [];

                        cartridges.forEach(function (cartridge) {
                            var profileJson = path.join(cartridgePath, cartridge, 'Data', 'Public', 'lib', cartridge.toLowerCase(), 'script-package-' + profile + '.json');

                            // load the JSON array from the *script-package-<profile>.json* if it exists
                            if (grunt.file.isFile(profileJson)) {
                                fs.readJsonSync(profileJson).forEach(function (module) {
                                    module = module.replace(/\./g, '/');
                                    // push module into package if not already there
                                    if (concatPackages[profile].indexOf(module) === -1) {
                                        concatPackages[profile].push(module);
                                    }
                                });
                            }
                        });

                        callback();
                    });
                });

                // perform the methods for getting the profiles with a parallel
                async.parallel(t_packages, callback);

            },

            create_packages = function (callback) {
                var t_packages = [];

                _.each(concatPackages, function (modules, packageName) {
                    // Create a task for each script package
                    t_packages.push(function (callback) {
                        var target = path.join(targetPath, 'package-' + packageName + '.js'),
                            packageInfo = 'var require=function(require){var metaparseConfig=require||(require={});metaparseConfig=require.config||(require.config={});metaparseConfig=metaparseConfig["jquery/metaparse"]||(metaparseConfig["jquery/metaparse"]={});metaparseConfig.autoparse={selector:metaparseConfig.autoparse?".ep-js":"#PREVENT-METAPARSE",require:epkStack=' + JSON.stringify(modules) + '};return require}(this.require);',

                            write_info_string = function (callback) {
                                // write info file
                                fs.writeFile(path.join(targetPath, 'package-' + packageName + '.info.js'), packageInfo, {mode: r_xr_____}, function (e) {
                                    if (e) { callback(e); }
                                    // write info string into package file
                                    fs.writeFile(target, packageInfo, {mode: r_xr_____}, callback);
                                });
                            },

                            write_package_file = function (callback) {
                                var t_module = [];

                                modules.forEach(function (module) {
                                    // Create a task for each module within the script package
                                    t_module.push(function (callback) {
                                        var content = commons.compileTemplateFiles(targetPath, path.join(targetPath, module + '.js'), false).concatenatedContent;
                                        fs.appendFile(target, content, callback);
                                    });
                                });

                                async.series(t_module, callback);
                            };

                        async.series([
                            write_info_string,
                            write_package_file
                        ], callback);
                    });
                });

                // perform the methods for writing the package files with a parallel
                async.parallel(t_packages, callback);
            };

        // Do the magic
        async.series([
            function (callback) {
                async.waterfall([
                    prepare,
                    get_script_packages
                ], callback);
            },
            create_packages
        // Set correct owership when series finished
        ], function (e) {
            if (_.isNumber(config.uid) && _.isNumber(config.gid)) {
                chownr(targetPath, config.uid, config.gid, callback);
            } else {
                callback(e);
            }
        });
    };

    // create packages for less/css files
    module.exports.less = function (config, callback) {
        var stylePackages,
            concatPackages = {},

            prepare = function (callback) {
                setTargetPath(config.webroot, config.version, config.store_type);

                // use all available stylePackages if no certain profile(s) is/are given
                stylePackages = config.profile ? config.profile.split(' ') : ['bo', 'sf', 'mobilesf', 'pageeditor', 'editor', 'testsuite'];

                // Set cartridge path and pass all cartridges to get_script_packages via waterfall callback
                getCartridgeData(config.store_type, callback);
            },

            get_style_packages = function (cartridgePath, cartridges, callback) {
                var t_packages = [];

                stylePackages.forEach(function (profile) {
                    // Create a task for each style package
                    t_packages.push(function (callback) {
                        var pendingFiles = [];

                        concatPackages[profile] = [];

                        // Push every path (css, less and variables file) separately to the corresponding profile array
                        cartridges.forEach(function (cartridge) {
                            var profileFile = path.join(targetPath, cartridge.toLowerCase(), 'style-package-' + profile);

                            if (grunt.file.isFile(profileFile + '-variables.less')) {
                                concatPackages[profile].push(profileFile + '-variables.less');
                            }

                            // Collect all other files before pushing them
                            pendingFiles.push.apply(
                                pendingFiles,
                                grunt.file.expand(profileFile + '.*')
                            );
                        });

                        // Push all other files after the variables
                        concatPackages[profile].push.apply(
                            concatPackages[profile],
                            pendingFiles
                        );

                        callback();
                    });
                });

                // Run tasks async
                async.series(t_packages, callback);
            },

            create_packages = function (callback) {
                async.each(stylePackages, function (packageName, callback) {
                    var target = path.join(targetPath, 'package-' + packageName + '.less'),
                        content = '';

                    concatPackages[packageName].forEach(function (p) {
                        p = commons.fixPath(path.relative(targetPath, p));
                        content += '@import url("' + p + '");\n';
                    });

                    fs.writeFile(target, content, {mode: r_xr_____}, callback);
                }, callback);
            };

        async.series([
            // Prerequisites
            function (callback) {
                async.waterfall([
                    prepare,
                    get_style_packages
                ], callback);
            },
            create_packages
        // Set correct owership when series finished
        ], function (e) {
            if (_.isNumber(config.uid) && _.isNumber(config.gid)) {
                chownr(targetPath, config.uid, config.gid, callback);
            } else {
                callback(e);
            }
        });
    };
}());