/*jslint node:true*/
/*global describe, after, before, it*/

(function () {

    'use strict';

    var async = require('async'),
        expect = require('expect.js'),
        fs = require('fs'),
        grunt = require('grunt'),
        sinon = require('sinon'),
        pack = require('../lib/pack'),
        path = require('path'),
        os = require('os'),
        perl = require('../lib/perl'),

        is_unixWithRoot = (os.platform() !== 'win32' && process.getuid() === 0) ? 'Running on Unix with root, running pack test with checks for uid and gid' : undefined;

    describe('Running pack-lib', function () {
        var cartridgePath = path.join(__dirname, 'fs/cartridges'),//commonsTesting.getPath('cartridges'),
            actualPath = path.join(__dirname, 'fs/webroot/StoreTypes/6.xx.x/MyStore/lib'), //commonsTesting.getPath('webroot-lib'),
            expectedPath= path.join(__dirname, 'fs/expected'), //commonsTesting.getPath('expected'),
            webrootPath = path.join(__dirname, 'fs/webroot'),
            options = {
                store_type: 'MyStore',
                version: '6.xx.x',
                webroot: webrootPath
            },
            stub;

        // invalidate require cache because we overwrote some arguments in the test for the argparser factory
        require.cache[path.normalize(__dirname + '/../lib/pack.js')] = undefined;

        before(function () {
            if (is_unixWithRoot) {
                options.uid = 1337;
                options.gid = 42;
            }
            // mock the perl module
            stub = sinon.stub(perl, 'get');

            stub.withArgs('cartridges', options.store_type).yields(null, [
                'PARENT/CartridgeA',
                'PARENT/CartridgeB',
                'Another_Parent/CartridgeC',
                'PARENT/CartridgeD',
                'PARENT/CartridgeE',
                'PARENT/JQuery'
            ]);

            stub.withArgs('env').yields(null, {cartridges: cartridgePath});
        });

        after(function () {
            perl.get.restore();
            grunt.file.expand([
                actualPath + '/package-*',
                '!' + actualPath + '/*-tominify*'
            ]).forEach(function (file) {
                if (grunt.file.isFile(file)) { grunt.file['delete'](file); }
            });
        });

        describe('for *.js files', function () {

            before(function (callback) {
                pack.js(options, callback);
            });

            it('should create all relevant package and info files', function () {
                expect(grunt.file.isFile(
                    path.join(actualPath, 'package-bo.js'),
                    path.join(actualPath, 'package-editor.js'),
                    path.join(actualPath, 'package-sf.js'),
                    path.join(actualPath, 'package-mobilesf.js'),
                    path.join(actualPath, 'package-sf_appointment.js'),
                    path.join(actualPath, 'package-testsuite.js'),
                    path.join(actualPath, 'package-bo.info.js'),
                    path.join(actualPath, 'package-editor.info.js'),
                    path.join(actualPath, 'package-sf.info.js'),
                    path.join(actualPath, 'package-mobilesf.info.js'),
                    path.join(actualPath, 'package-sf_appointment.info.js'),
                    path.join(actualPath, 'package-testsuite.info.js')
                ));
            });

            it('should create correct package-bo.js', function (callback) {
                async.parallel({
                    expected: function (callback) {
                        fs.readFile(path.join(expectedPath, 'package-bo.js'), callback);
                    },

                    actual: function (callback) {
                        fs.readFile(path.join(actualPath, 'package-bo.js'), callback);
                    }
                }, function (error, result) {
                    expect(error).to.not.be.ok();
                    expect(result.expected.toString()).to.eql(result.actual.toString());

                    // if we're root, we check if file ownership is set correctly
                    if (is_unixWithRoot) {
                        // check uid, gid and permissions
                        fs.stat(path.join(actualPath, 'package-bo.js'), function (err, stats) {
                            expect(err).to.not.be.ok();

                            expect(stats.uid).to.be.a('number');
                            expect(stats.uid).to.be(1337);
                            expect(stats.gid).to.be.a('number');
                            expect(stats.gid).to.be(42);

                            expect(parseInt(stats.mode.toString(8), 10)).to.be(100640);

                            callback();
                        });
                    } else {
                        callback();
                    }
                });
            });

            it('should create correct package-bo.info.js', function (callback) {
                async.parallel({
                    expected: function (callback) {
                        fs.readFile(path.join(expectedPath, 'package-bo.info.js'), callback);
                    },

                    actual: function (callback) {
                        fs.readFile(path.join(actualPath, 'package-bo.info.js'), callback);
                    }
                }, function (error, result) {
                    expect(error).to.not.be.ok();
                    expect(result.expected.toString()).to.eql(result.actual.toString());

                    // if we're root, we check if file ownership is set correctly
                    if (is_unixWithRoot) {
                        // check uid, gid and permissions
                        fs.stat(path.join(actualPath, 'package-bo.info.js'), function (err, stats) {
                            expect(err).to.not.be.ok();

                            expect(stats.uid).to.be.a('number');
                            expect(stats.uid).to.be(1337);
                            expect(stats.gid).to.be.a('number');
                            expect(stats.gid).to.be(42);

                            expect(parseInt(stats.mode.toString(8), 10)).to.be(100640);

                            callback();
                        });
                    } else {
                        callback();
                    }
                });
            });
        });

        describe('for *.less files', function () {

            before(function (callback) {
                pack.less(options, callback);
            });

            it('should create all relevant package files', function () {
                expect(grunt.file.isFile(
                    path.join(actualPath, 'package-bo.less'),
                    path.join(actualPath, 'package-editor.less'),
                    path.join(actualPath, 'package-sf.less'),
                    path.join(actualPath, 'package-mobilesf.less'),
                    path.join(actualPath, 'package-pageeditor.less'),
                    path.join(actualPath, 'package-testsuite.less')
                ));
            });

            it('should create correct package-bo.less', function (callback) {
                async.parallel({
                    expected: function (callback) {
                        fs.readFile(path.join(expectedPath, 'package-bo.less'), callback);
                    },

                    actual: function (callback) {
                        fs.readFile(path.join(actualPath, 'package-bo.less'), callback);
                    }
                }, function (error, result) {
                    expect(error).to.not.be.ok();
                    // replace CR (\r) and LF (\n) by LF (\n) due to different new line handling in different operating systems
                    expect(result.expected.toString().replace(/\r\n/g, "\n")).to.be(result.actual.toString().replace(/\r\n/g, "\n"));

                    // if we're root, we check if file ownership is set correctly
                    if (is_unixWithRoot) {
                        // check uid, gid and permissions
                        fs.stat(path.join(actualPath, 'package-bo.less'), function (err, stats) {
                            expect(err).to.not.be.ok();

                            expect(stats.uid).to.be.a('number');
                            expect(stats.uid).to.be(1337);
                            expect(stats.gid).to.be.a('number');
                            expect(stats.gid).to.be(42);

                            expect(parseInt(stats.mode.toString(8), 10)).to.be(100640);

                            callback();
                        });
                    } else {
                        callback();
                    }
                });
            });
        });
    });
}());
